#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "matrix.h"

/* Genera una stringa random */
static int generate_random_string(int *size, char **value);
/* Comparatore per due elementi di una colonna */
static int cols_cmp(const void *a, const void *b);
/* Comparatore per due righe */
static int rows_cmp(const void *a, const void *b);

/* Ritorna un numero (random) tra 0 e (max-1) */
#define RAND(max) ((int) ((max) * (rand() / (RAND_MAX + 1.0))))

int matrix_generate(matrix *matrix) {
  int i;

  /* Alloca la memoria per un array di 'row' */
  if (!(matrix->m = calloc(matrix->rows, sizeof(row)))) {
	perror("Allocazione della memoria (row)");
	return 1;
  }	

  /* Per ogni riga, alloca la memoria per un array di 'node' */
  for (i = 0; i < matrix->rows; i++) {
	if (!(matrix->m[i].cols = calloc(matrix->cols, sizeof(node)))) {
	  perror("Allocazione della memoria (cols)");
	  return 1;
	}
  }
  
  return 0;
}

int matrix_populate(matrix *matrix) {
  int i, j;
  node *node;

  /* Imposta il seme */
  srand((unsigned int) time(NULL));
  
  for (i = 0; i < matrix->rows; i++) {
	/* Imposta il rank a zero */
	matrix->m[i].rank = 0;
	for (j = 0; j < matrix->cols; j++) {
	  node = &(matrix->m[i].cols[j]);
	  /* Genera la stringa random */
	  if (generate_random_string(&(node->size), &(node->value))) {
		return 1;
	  }
	  /* Incrementa il rank della riga */
	  matrix->m[i].rank += node->size;
	}
  }

  return 0;
}

int matrix_dispose(matrix *matrix) {
  int i, j;
  
  /* Rilascia tutta la memoria allocata */
  for (i = 0; i < matrix->rows; i++) {
	for (j = 0; j < matrix->cols; j++) {
	  if (matrix->m[i].cols[j].value) {
		free(matrix->m[i].cols[j].value);
	  }
	}
	free(matrix->m[i].cols);
  }
  free(matrix->m);
  
  return 0;
}

void matrix_print(matrix *matrix, const char *title) {
  int i, j;
  node *node;
  
  printf("========= %s =========\n\n", title);
  for (i = 0; i < matrix->rows; i++) {
	printf("### Row %d [rank: %d] ###\n", i+1, matrix->m[i].rank);
	for (j = 0; j < matrix->cols; j++) {
	  node = &(matrix->m[i].cols[j]);
	  printf(" [%d]: %s (size: %d)\n", j+1, node->value, node->size);
	}
	printf("\n");
  }
}

static int generate_random_string(int *size, char **value) {
  int i;
  char *str;

  *size = RAND(MAX_VALUE_LENGTH) + 1;
  if (*size) {
	if (!(str = malloc(sizeof(char) * ((*size) + 1)))) {
	  perror("Allocating memory for one string");
	  return 1;
	}
	*value = str;
	for (i = 0; i < *size; i++, str++) {
	  *str = 'a' + (RAND('z' - 'a'));
	}
	*str = '\0';
  }

  return 0;
}

void matrix_sort(matrix *matrix) {
  int i;

  for (i = 0; i < matrix->rows; i++) {
	/* Ordina le colonne di una riga in base alle stringhe alfanumeriche */
	qsort(matrix->m[i].cols, matrix->cols, sizeof(node), cols_cmp);
  }

  /* Ordina le righe della matrice in base al rank */
  qsort(matrix->m, matrix->rows, sizeof(row), rows_cmp);
}

static int cols_cmp(const void *a, const void *b) {
  const node *n1, *n2;

  n1 = (const node *) a;
  n2 = (const node *) b;
  
  return strcmp(n1->value, n2->value);
}

static int rows_cmp(const void *a, const void *b) {
  const row *n1, *n2;

  n1 = (const row *) a;
  n2 = (const row *) b;

  if (n1->rank < n2->rank) {
	return -1;
  }

  if (n1->rank > n2->rank) {
	return 1;
  }

  return 0;
}
